@@ -17,15 +17,19 @@
 from glob import glob
 import os, subprocess, shutil, sys, signal
 from io import StringIO
-import sys
+from ast import literal_eval
+import sys, tempfile
 from mesonbuild import environment
 from mesonbuild import mesonlib
 from mesonbuild import mlog
 from mesonbuild import mesonmain
+from mesonbuild.mesonlib import stringlistify
 from mesonbuild.scripts import meson_test, meson_benchmark
 import argparse
 import xml.etree.ElementTree as ET
 import time
+import multiprocessing
+import concurrent.futures as conc
 
 from mesonbuild.mesonmain import backendlist
 
@@ -38,14 +42,36 @@ class TestResult:
         self.buildtime = buildtime
         self.testtime = testtime
 
+class AutoDeletedDir():
+    def __init__(self, dir):
+        self.dir = dir
+    def __enter__(self):
+        os.makedirs(self.dir, exist_ok=True)
+        return self.dir
+    def __exit__(self, type, value, traceback):
+        # On Windows, shutil.rmtree fails sometimes, because 'the directory is not empty'.
+        # Retrying fixes this.
+        # That's why we don't use tempfile.TemporaryDirectory, but wrap the deletion in the AutoDeletedDir class.
+        retries = 5
+        for i in range(0, retries):
+            try:
+                shutil.rmtree(self.dir)
+                return
+            except OSError:
+                if i == retries-1:
+                    raise
+                time.sleep(0.1 * (2**i))
+
 passing_tests = 0
 failing_tests = 0
 skipped_tests = 0
 print_debug = 'MESON_PRINT_TEST_OUTPUT' in os.environ
 
-test_build_dir = 'work area'
-install_dir = os.path.join(os.path.split(os.path.abspath(__file__))[0], 'install dir')
 meson_command = os.path.join(os.getcwd(), 'meson')
+if not os.path.exists(meson_command):
+    meson_command += '.py'
+    if not os.path.exists(meson_command):
+        raise RuntimeError('Could not find main Meson script to run.')
 
 class StopException(Exception):
     def __init__(self):
@@ -130,14 +156,8 @@ def validate_install(srcdir, installdir):
         return 'Found extra file %s.' % fname
     return ''
 
-def log_text_file(logfile, testdir, msg, stdo, stde):
-    global passing_tests, failing_tests, stop
-    if msg != '':
-        print('Fail:', msg)
-        failing_tests += 1
-    else:
-        print('Success')
-        passing_tests += 1
+def log_text_file(logfile, testdir, stdo, stde):
+    global stop
     logfile.write('%s\nstdout\n\n---\n' % testdir)
     logfile.write(stdo)
     logfile.write('\n\n---\n\nstderr\n\n---\n')
@@ -177,18 +197,35 @@ def run_test_inprocess(testdir):
         os.chdir(old_cwd)
     return (max(returncode_test, returncode_benchmark), mystdout.getvalue(), mystderr.getvalue())
 
+def parse_test_args(testdir):
+    args = []
+    try:
+        with open(os.path.join(testdir, 'test_args.txt'), 'r') as f:
+            content = f.read()
+            try:
+                args = literal_eval(content)
+            except Exception:
+                raise Exception('Malformed test_args file.')
+            args = stringlistify(args)
+    except FileNotFoundError:
+        pass
+    return args
 
-def run_test(testdir, extra_args, should_succeed):
-    global compile_commands
-    mlog.shutdown() # Close the log file because otherwise Windows wets itself.
-    shutil.rmtree(test_build_dir)
-    shutil.rmtree(install_dir)
-    os.mkdir(test_build_dir)
-    os.mkdir(install_dir)
-    print('Running test: ' + testdir)
+def run_test(skipped, testdir, extra_args, flags, compile_commands, install_commands, should_succeed):
+    if skipped:
+        return None
+    with AutoDeletedDir(tempfile.mkdtemp(prefix='b ', dir='.')) as build_dir:
+        with AutoDeletedDir(tempfile.mkdtemp(prefix='i ', dir=os.getcwd())) as install_dir:
+            try:
+                return _run_test(testdir, build_dir, install_dir, extra_args, flags, compile_commands, install_commands, should_succeed)
+            finally:
+                mlog.shutdown() # Close the log file because otherwise Windows wets itself.
+
+def _run_test(testdir, test_build_dir, install_dir, extra_args, flags, compile_commands, install_commands, should_succeed):
+    test_args = parse_test_args(testdir)
     gen_start = time.time()
     gen_command = [meson_command, '--prefix', '/usr', '--libdir', 'lib', testdir, test_build_dir]\
-        + unity_flags + backend_flags + extra_args
+        + flags + test_args + extra_args
     (returncode, stdo, stde) = run_configure_inprocess(gen_command)
     gen_time = time.time() - gen_start
     if not should_succeed:
@@ -207,8 +244,8 @@ def run_test(testdir, extra_args, should_succeed):
                           stdout=subprocess.PIPE, stderr=subprocess.PIPE)
     (o, e) = pc.communicate()
     build_time = time.time() - build_start
-    stdo += o.decode('utf-8')
-    stde += e.decode('utf-8')
+    stdo += o.decode(sys.stdout.encoding)
+    stde += e.decode(sys.stdout.encoding)
     if pc.returncode != 0:
         return TestResult('Compiling source code failed.', stdo, stde, gen_time, build_time)
     test_start = time.time()
@@ -222,7 +259,6 @@ def run_test(testdir, extra_args, should_succeed):
     if returncode != 0:
         return TestResult('Running unit tests failed.', stdo, stde, gen_time, build_time, test_time)
     if len(install_commands) == 0:
-        print("Skipping install test")
         return TestResult('', '', '', gen_time, build_time, test_time)
     else:
         env = os.environ.copy()
@@ -230,8 +266,8 @@ def run_test(testdir, extra_args, should_succeed):
         pi = subprocess.Popen(install_commands, cwd=test_build_dir, env=env,
                               stdout=subprocess.PIPE, stderr=subprocess.PIPE)
         (o, e) = pi.communicate()
-        stdo += o.decode('utf-8')
-        stde += e.decode('utf-8')
+        stdo += o.decode(sys.stdout.encoding)
+        stde += e.decode(sys.stdout.encoding)
         if pi.returncode != 0:
             return TestResult('Running install failed.', stdo, stde, gen_time, build_time, test_time)
         return TestResult(validate_install(testdir, install_dir), stdo, stde, gen_time, build_time, test_time)
@@ -260,23 +296,19 @@ def detect_tests_to_run():
     all_tests.append(('objective c', gather_tests('test cases/objc'), False if not mesonlib.is_windows() else True))
     all_tests.append(('fortran', gather_tests('test cases/fortran'), False if shutil.which('gfortran') else True))
     all_tests.append(('swift', gather_tests('test cases/swift'), False if shutil.which('swiftc') else True))
+    all_tests.append(('python3', gather_tests('test cases/python3'), False if shutil.which('python3') else True))
     return all_tests
 
 def run_tests(extra_args):
+    global passing_tests, failing_tests, stop
     all_tests = detect_tests_to_run()
     logfile = open('meson-test-run.txt', 'w', encoding="utf_8")
     junit_root = ET.Element('testsuites')
     conf_time = 0
     build_time = 0
     test_time = 0
-    try:
-        os.mkdir(test_build_dir)
-    except OSError:
-        pass
-    try:
-        os.mkdir(install_dir)
-    except OSError:
-        pass
+
+    executor = conc.ProcessPoolExecutor(max_workers=multiprocessing.cpu_count())
 
     for name, test_cases, skipped in all_tests:
         current_suite = ET.SubElement(junit_root, 'testsuite', {'name' : name, 'tests' : str(len(test_cases))})
@@ -284,28 +316,40 @@ def run_tests(extra_args):
             print('\nNot running %s tests.\n' % name)
         else:
             print('\nRunning %s tests.\n' % name)
+        futures = []
         for t in test_cases:
             # Jenkins screws us over by automatically sorting test cases by name
             # and getting it wrong by not doing logical number sorting.
             (testnum, testbase) = os.path.split(t)[-1].split(' ', 1)
             testname = '%.3d %s' % (int(testnum), testbase)
-            if skipped:
+            result = executor.submit(run_test, skipped, t, extra_args, unity_flags + backend_flags, compile_commands, install_commands, name != 'failing')
+            futures.append((testname, t, result))
+        for (testname, t, result) in futures:
+            result = result.result()
+            if result is None:
+                print('Skipping:', t)
                 current_test = ET.SubElement(current_suite, 'testcase', {'name' : testname,
                                                                          'classname' : name})
                 ET.SubElement(current_test, 'skipped', {})
                 global skipped_tests
                 skipped_tests += 1
             else:
-                ts = time.time()
-                result = run_test(t, extra_args, name != 'failing')
-                te = time.time()
+                without_install = "" if len(install_commands) > 0 else " (without install)"
+                if result.msg != '':
+                    print('Failed test%s: %s' % (without_install, t))
+                    print('Reason:', result.msg)
+                    failing_tests += 1
+                else:
+                    print('Succeeded test%s: %s' % (without_install, t))
+                    passing_tests += 1
                 conf_time += result.conftime
                 build_time += result.buildtime
                 test_time += result.testtime
-                log_text_file(logfile, t, result.msg, result.stdo, result.stde)
+                total_time = conf_time + build_time + test_time
+                log_text_file(logfile, t, result.stdo, result.stde)
                 current_test = ET.SubElement(current_suite, 'testcase', {'name' : testname,
                                                                          'classname' : name,
-                                                                         'time' : '%.3f' % (te - ts)})
+                                                                         'time' : '%.3f' % total_time})
                 if result.msg != '':
                     ET.SubElement(current_test, 'failure', {'message' : result.msg})
                 stdoel = ET.SubElement(current_test, 'system-out')
@@ -331,7 +375,7 @@ def check_file(fname):
 def check_format():
     for (root, _, files) in os.walk('.'):
         for file in files:
-            if file.endswith('.py') or file.endswith('.build'):
+            if file.endswith('.py') or file.endswith('.build') or file == 'meson_options.txt':
                 fullname = os.path.join(root, file)
                 check_file(fullname)
 
